@@ -59,8 +59,6 @@ static void ovirt_foreign_menu_fetch_vm_cdrom_async(OvirtForeignMenu *menu, GTas
 static void ovirt_foreign_menu_refresh_cdrom_file_async(OvirtForeignMenu *menu, GTask *task);
 static void ovirt_foreign_menu_fetch_iso_list_async(OvirtForeignMenu *menu, GTask *task);
 
-G_DEFINE_TYPE (OvirtForeignMenu, ovirt_foreign_menu, G_TYPE_OBJECT)
-
 
 struct _OvirtForeignMenuPrivate {
     OvirtProxy *proxy;
@@ -87,7 +85,7 @@ struct _OvirtForeignMenuPrivate {
 };
 
 
-#define OVIRT_FOREIGN_MENU_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), OVIRT_TYPE_FOREIGN_MENU, OvirtForeignMenuPrivate))
+G_DEFINE_TYPE_WITH_PRIVATE (OvirtForeignMenu, ovirt_foreign_menu, G_TYPE_OBJECT)
 
 
 enum {
@@ -100,7 +98,6 @@ enum {
     PROP_VM_GUID,
 };
 
-
 gchar *
 ovirt_foreign_menu_get_current_iso_name(OvirtForeignMenu *foreign_menu)
 {
@@ -230,8 +227,6 @@ ovirt_foreign_menu_class_init(OvirtForeignMenuClass *klass)
     oclass->set_property = ovirt_foreign_menu_set_property;
     oclass->dispose = ovirt_foreign_menu_dispose;
 
-    g_type_class_add_private(klass, sizeof(OvirtForeignMenuPrivate));
-
     g_object_class_install_property(oclass,
                                     PROP_PROXY,
                                     g_param_spec_object("proxy",
@@ -287,7 +282,7 @@ ovirt_foreign_menu_class_init(OvirtForeignMenuClass *klass)
 static void
 ovirt_foreign_menu_init(OvirtForeignMenu *self)
 {
-    self->priv = OVIRT_FOREIGN_MENU_GET_PRIVATE(self);
+    self->priv = ovirt_foreign_menu_get_instance_private(self);
 }
 
 
@@ -368,7 +363,7 @@ ovirt_foreign_menu_next_async_step(OvirtForeignMenu *menu,
     default:
         g_warn_if_reached();
         g_task_return_new_error(task, OVIRT_ERROR, OVIRT_ERROR_FAILED,
-                                "Invalid state: %d", current_state);
+                                "Invalid state: %u", current_state);
         g_object_unref(task);
     }
 }
@@ -490,7 +485,8 @@ static void ovirt_foreign_menu_set_files(OvirtForeignMenu *menu,
          * associated with file resources , but as of 3.2, this node
          * is not present, so we do an extension check instead
          * to differentiate between ISOs and floppy images */
-        if (g_str_has_suffix(name, ".vfd")) {
+        if (!g_str_has_suffix(name, ".iso")) {
+            g_debug("Ignoring %s which does not have a .iso extension", name);
             g_free(name);
             continue;
         }
@@ -623,6 +619,13 @@ static void ovirt_foreign_menu_fetch_vm_cdrom_async(OvirtForeignMenu *menu,
 }
 
 #ifdef HAVE_OVIRT_DATA_CENTER
+static gboolean strv_contains(const gchar * const *strv, const gchar *str)
+{
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+  return g_strv_contains (strv, str);
+G_GNUC_END_IGNORE_DEPRECATIONS
+}
+
 static gboolean storage_domain_attached_to_data_center(OvirtStorageDomain *domain,
                                                       OvirtDataCenter *data_center)
 {
@@ -632,7 +635,7 @@ static gboolean storage_domain_attached_to_data_center(OvirtStorageDomain *domai
 
     g_object_get(domain, "data-center-ids", &data_center_ids, NULL);
     g_object_get(data_center, "guid", &data_center_guid, NULL);
-    match = g_strv_contains((const gchar * const *) data_center_ids, data_center_guid);
+    match = strv_contains((const gchar * const *) data_center_ids, data_center_guid);
     g_strfreev(data_center_ids);
     g_free(data_center_guid);
 
@@ -640,6 +643,35 @@ static gboolean storage_domain_attached_to_data_center(OvirtStorageDomain *domai
 }
 #endif
 
+static gboolean storage_domain_validate(OvirtForeignMenu *menu G_GNUC_UNUSED,
+                                        OvirtStorageDomain *domain)
+{
+    char *name;
+    int type, state;
+    gboolean ret = TRUE;
+
+    g_object_get(domain, "name", &name, "type", &type, "state", &state, NULL);
+
+    if (type != OVIRT_STORAGE_DOMAIN_TYPE_ISO) {
+        g_debug("Storage domain '%s' type is not ISO", name);
+        ret = FALSE;
+    }
+
+    if (state != OVIRT_STORAGE_DOMAIN_STATE_ACTIVE) {
+        g_debug("Storage domain '%s' state is not active", name);
+        ret = FALSE;
+    }
+
+#ifdef HAVE_OVIRT_DATA_CENTER
+    if (!storage_domain_attached_to_data_center(domain, menu->priv->data_center)) {
+        g_debug("Storage domain '%s' is not attached to data center", name);
+        ret = FALSE;
+    }
+#endif
+
+    g_free(name);
+    return ret;
+}
 
 static void storage_domains_fetched_cb(GObject *source_object,
                                        GAsyncResult *result,
@@ -663,23 +695,9 @@ static void storage_domains_fetched_cb(GObject *source_object,
     g_hash_table_iter_init(&iter, ovirt_collection_get_resources(collection));
     while (g_hash_table_iter_next(&iter, NULL, (gpointer *)&domain)) {
         OvirtCollection *file_collection;
-        int type;
-        int state;
 
-        g_object_get(domain, "type", &type, "state", &state, NULL);
-        if (type != OVIRT_STORAGE_DOMAIN_TYPE_ISO) {
+        if (!storage_domain_validate(menu, domain))
             continue;
-        }
-
-        if (state != OVIRT_STORAGE_DOMAIN_STATE_ACTIVE) {
-            continue;
-        }
-
-#ifdef HAVE_OVIRT_DATA_CENTER
-        if (!storage_domain_attached_to_data_center(domain, menu->priv->data_center)) {
-            continue;
-        }
-#endif
 
         file_collection = ovirt_storage_domain_get_files(domain);
         if (file_collection != NULL) {
