@@ -54,9 +54,7 @@ struct _RemoteViewerPrivate {
     gboolean open_recent_dialog;
 };
 
-G_DEFINE_TYPE (RemoteViewer, remote_viewer, VIRT_VIEWER_TYPE_APP)
-#define GET_PRIVATE(o)                                                        \
-    (G_TYPE_INSTANCE_GET_PRIVATE ((o), REMOTE_VIEWER_TYPE, RemoteViewerPrivate))
+G_DEFINE_TYPE_WITH_PRIVATE (RemoteViewer, remote_viewer, VIRT_VIEWER_TYPE_APP)
 
 enum RemoteViewerProperties {
     PROP_0,
@@ -196,8 +194,6 @@ remote_viewer_class_init (RemoteViewerClass *klass)
     VirtViewerAppClass *app_class = VIRT_VIEWER_APP_CLASS (klass);
     GApplicationClass *g_app_class = G_APPLICATION_CLASS(klass);
 
-    g_type_class_add_private (klass, sizeof (RemoteViewerPrivate));
-
     object_class->get_property = remote_viewer_get_property;
     object_class->dispose = remote_viewer_dispose;
 
@@ -221,7 +217,7 @@ remote_viewer_class_init (RemoteViewerClass *klass)
 static void
 remote_viewer_init(RemoteViewer *self)
 {
-    self->priv = GET_PRIVATE(self);
+    self->priv = remote_viewer_get_instance_private(self);
 }
 
 RemoteViewer *
@@ -345,10 +341,8 @@ authenticate_cb(RestProxy *proxy, G_GNUC_UNUSED RestProxyAuth *auth,
                      "username", username,
                      "password", password,
                      NULL);
-#ifdef HAVE_OVIRT_CANCEL
     } else {
         rest_proxy_auth_cancel(auth);
-#endif
     }
 
     g_free(username);
@@ -447,14 +441,12 @@ create_ovirt_session(VirtViewerApp *app, const char *uri, GError **err)
     api = ovirt_proxy_fetch_api(proxy, &error);
     if (error != NULL) {
         g_debug("failed to get oVirt 'api' collection: %s", error->message);
-#ifdef HAVE_OVIRT_CANCEL
         if (g_error_matches(error, OVIRT_REST_CALL_ERROR, OVIRT_REST_CALL_ERROR_CANCELLED)) {
             g_clear_error(&error);
             g_set_error_literal(&error,
                                 VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_CANCELLED,
                                 _("Authentication was cancelled"));
         }
-#endif
         goto error;
     }
     vms = ovirt_api_get_vms(api);
@@ -509,10 +501,10 @@ create_ovirt_session(VirtViewerApp *app, const char *uri, GError **err)
                  "proxy-url", &proxy_url,
                  NULL);
     if (port != 0) {
-        gport = g_strdup_printf("%d", port);
+        gport = g_strdup_printf("%u", port);
     }
     if (secure_port != 0) {
-        gtlsport = g_strdup_printf("%d", secure_port);
+        gtlsport = g_strdup_printf("%u", secure_port);
     }
 
     if (ghost == NULL) {
@@ -528,7 +520,7 @@ create_ovirt_session(VirtViewerApp *app, const char *uri, GError **err)
         session_type = "vnc";
     } else {
         g_set_error(&error, VIRT_VIEWER_ERROR, VIRT_VIEWER_ERROR_FAILED,
-                    _("oVirt VM %s has unknown display type: %d"), vm_name, type);
+                    _("oVirt VM %s has unknown display type: %u"), vm_name, type);
         g_debug("%s", error->message);
         goto error;
     }
@@ -688,9 +680,18 @@ remote_viewer_initial_connect(RemoteViewer *self, const gchar *type, const gchar
 {
     VirtViewerApp *app = VIRT_VIEWER_APP(self);
 
-    if (!virt_viewer_app_create_session(app, type, error))
-        return FALSE;
-
+#ifdef HAVE_OVIRT
+    if (g_strcmp0(type, "ovirt") == 0) {
+        if (!create_ovirt_session(app, guri, error)) {
+            g_prefix_error(error, _("Couldn't open oVirt session: "));
+            return FALSE;
+        }
+    } else
+#endif
+    {
+        if (!virt_viewer_app_create_session(app, type, error))
+            return FALSE;
+    }
 
     g_signal_connect(virt_viewer_app_get_session(app), "session-connected",
                      G_CALLBACK(remote_viewer_session_connected), g_strdup(guri));
@@ -786,18 +787,8 @@ retry_dialog:
                                 _("Cannot determine the connection type from URI"));
             goto cleanup;
         }
-#ifdef HAVE_OVIRT
-        if (g_strcmp0(type, "ovirt") == 0) {
-            if (!create_ovirt_session(app, guri, &error)) {
-                g_prefix_error(&error, _("Couldn't open oVirt session: "));
-                goto cleanup;
-            }
-        } else
-#endif
-        {
-            if (!remote_viewer_initial_connect(self, type, guri, vvfile, &error))
-                goto cleanup;
-        }
+        if (!remote_viewer_initial_connect(self, type, guri, vvfile, &error))
+            goto cleanup;
     }
 
     ret = VIRT_VIEWER_APP_CLASS(remote_viewer_parent_class)->start(app, &error);
