@@ -62,6 +62,8 @@
 #include "virt-viewer-session-spice.h"
 #endif
 
+#include "virt-viewer-display-vte.h"
+
 gboolean doDebug = FALSE;
 
 /* Signal handlers for about dialog */
@@ -110,7 +112,7 @@ struct _VirtViewerAppPrivate {
     VirtViewerWindow *main_window;
     GtkWidget *main_notebook;
     GList *windows;
-    GHashTable *displays;
+    GHashTable *displays; /* !vte */
     GHashTable *initial_display_map;
     gchar *clipboard;
     GtkWidget *preferences;
@@ -125,6 +127,8 @@ struct _VirtViewerAppPrivate {
     gboolean attach;
     gboolean quitting;
     gboolean kiosk;
+    gboolean vm_ui;
+    gboolean vm_running;
 
     VirtViewerSession *session;
     gboolean active;
@@ -145,7 +149,6 @@ struct _VirtViewerAppPrivate {
     char *title;
     char *uuid;
 
-    gint focused;
     GKeyFile *config;
     gchar *config_file;
 
@@ -157,9 +160,7 @@ struct _VirtViewerAppPrivate {
 };
 
 
-G_DEFINE_ABSTRACT_TYPE(VirtViewerApp, virt_viewer_app, GTK_TYPE_APPLICATION)
-#define GET_PRIVATE(o)                                                        \
-    (G_TYPE_INSTANCE_GET_PRIVATE ((o), VIRT_VIEWER_TYPE_APP, VirtViewerAppPrivate))
+G_DEFINE_ABSTRACT_TYPE_WITH_PRIVATE(VirtViewerApp, virt_viewer_app, GTK_TYPE_APPLICATION)
 
 enum {
     PROP_0,
@@ -170,10 +171,11 @@ enum {
     PROP_FULLSCREEN,
     PROP_TITLE,
     PROP_ENABLE_ACCEL,
-    PROP_HAS_FOCUS,
     PROP_KIOSK,
     PROP_QUIT_ON_DISCONNECT,
     PROP_UUID,
+    PROP_VM_UI,
+    PROP_VM_RUNNING,
 };
 
 void
@@ -258,9 +260,13 @@ virt_viewer_app_save_config(VirtViewerApp *self)
         if (error) {
             g_debug("Unable to get comment from key file: %s", error->message);
             g_clear_error(&error);
-        } else {
-            if (!comment || *comment == '\0')
-                g_key_file_set_comment(priv->config, priv->uuid, NULL, priv->guest_name, NULL);
+        }
+
+        if (comment == NULL ||
+                (comment != NULL && g_strstr_len(comment, -1, priv->guest_name) == NULL)) {
+            /* Note that this function appends the guest's name string as last
+             * comment in case there were comments there already */
+            g_key_file_set_comment(priv->config, priv->uuid, NULL, priv->guest_name, NULL);
         }
         g_free(comment);
     }
@@ -282,6 +288,11 @@ virt_viewer_app_quit(VirtViewerApp *self)
 
     virt_viewer_app_save_config(self);
 
+    if (priv->vm_ui) {
+        virt_viewer_session_vm_action(VIRT_VIEWER_SESSION(priv->session),
+                                      VIRT_VIEWER_SESSION_VM_ACTION_QUIT);
+    }
+
     if (priv->session) {
         virt_viewer_session_close(VIRT_VIEWER_SESSION(priv->session));
         if (priv->connected) {
@@ -294,7 +305,7 @@ virt_viewer_app_quit(VirtViewerApp *self)
 }
 
 static gint
-get_n_client_monitors()
+get_n_client_monitors(void)
 {
     return gdk_screen_get_n_monitors(gdk_screen_get_default());
 }
@@ -502,29 +513,6 @@ virt_viewer_app_get_n_windows_visible(VirtViewerApp *self)
     return n;
 }
 
-gboolean
-virt_viewer_app_window_set_visible(VirtViewerApp *self,
-                                   VirtViewerWindow *window,
-                                   gboolean visible)
-{
-    g_return_val_if_fail(VIRT_VIEWER_IS_APP(self), FALSE);
-    g_return_val_if_fail(VIRT_VIEWER_IS_WINDOW(window), FALSE);
-
-    if (visible) {
-        virt_viewer_window_show(window);
-        return TRUE;
-    } else {
-        if (virt_viewer_app_get_n_windows_visible(self) > 1) {
-            virt_viewer_window_hide(window);
-            return FALSE;
-        }
-
-        virt_viewer_app_maybe_quit(self, window);
-    }
-
-    return FALSE;
-}
-
 static void hide_one_window(gpointer value,
                             gpointer user_data G_GNUC_UNUSED)
 {
@@ -724,13 +712,23 @@ virt_viewer_app_set_window_subtitle(VirtViewerApp *app,
     const gchar *title = virt_viewer_app_get_title(app);
 
     if (title != NULL) {
+        VirtViewerDisplay *display = virt_viewer_window_get_display(window);
         gchar *d = strstr(title, "%d");
+        gchar *desc = NULL;
+
+        if (display && VIRT_VIEWER_IS_DISPLAY_VTE(display)) {
+            g_object_get(display, "name", &desc, NULL);
+        } else  {
+            desc = g_strdup_printf("%d", nth + 1);
+        }
+
         if (d != NULL) {
             *d = '\0';
-            subtitle = g_strdup_printf("%s%d%s", title, nth + 1, d + 2);
+            subtitle = g_strdup_printf("%s%s%s", title, desc, d + 2);
             *d = '%';
         } else
-            subtitle = g_strdup_printf("%s (%d)", title, nth + 1);
+            subtitle = g_strdup_printf("%s (%s)", title, desc);
+        g_free(desc);
     }
 
     g_object_set(window, "subtitle", subtitle, NULL);
@@ -738,8 +736,8 @@ virt_viewer_app_set_window_subtitle(VirtViewerApp *app,
 }
 
 static void
-set_title(gpointer value,
-          gpointer user_data)
+set_subtitle(gpointer value,
+             gpointer user_data)
 {
     VirtViewerApp *app = user_data;
     VirtViewerWindow *window = value;
@@ -755,7 +753,7 @@ set_title(gpointer value,
 static void
 virt_viewer_app_set_all_window_subtitles(VirtViewerApp *app)
 {
-    g_list_foreach(app->priv->windows, set_title, app);
+    g_list_foreach(app->priv->windows, set_subtitle, app);
 }
 
 static void update_title(gpointer value,
@@ -801,6 +799,10 @@ static VirtViewerWindow *
 virt_viewer_app_get_nth_window(VirtViewerApp *self, gint nth)
 {
     GList *l;
+
+    if (nth < 0)
+        return NULL;
+
     for (l = self->priv->windows; l; l = l->next) {
         VirtViewerDisplay *display = virt_viewer_window_get_display(l->data);
         if (display
@@ -818,33 +820,6 @@ viewer_window_visible_cb(GtkWidget *widget G_GNUC_UNUSED,
     virt_viewer_app_update_menu_displays(VIRT_VIEWER_APP(user_data));
 }
 
-static gboolean
-viewer_window_focus_in_cb(GtkWindow *window G_GNUC_UNUSED,
-                          GdkEvent *event G_GNUC_UNUSED,
-                          VirtViewerApp *self)
-{
-    self->priv->focused += 1;
-
-    if (self->priv->focused == 1)
-        g_object_notify(G_OBJECT(self), "has-focus");
-
-    return FALSE;
-}
-
-static gboolean
-viewer_window_focus_out_cb(GtkWindow *window G_GNUC_UNUSED,
-                           GdkEvent *event G_GNUC_UNUSED,
-                           VirtViewerApp *self)
-{
-    self->priv->focused -= 1;
-    g_warn_if_fail(self->priv->focused >= 0);
-
-    if (self->priv->focused <= 0)
-        g_object_notify(G_OBJECT(self), "has-focus");
-
-    return FALSE;
-}
-
 static gboolean
 virt_viewer_app_has_usbredir(VirtViewerApp *self)
 {
@@ -881,16 +856,27 @@ virt_viewer_app_window_new(VirtViewerApp *self, gint nth)
 
     g_signal_connect(w, "hide", G_CALLBACK(viewer_window_visible_cb), self);
     g_signal_connect(w, "show", G_CALLBACK(viewer_window_visible_cb), self);
-    g_signal_connect(w, "focus-in-event", G_CALLBACK(viewer_window_focus_in_cb), self);
-    g_signal_connect(w, "focus-out-event", G_CALLBACK(viewer_window_focus_out_cb), self);
     return window;
 }
 
+static void
+window_weak_notify(gpointer data, GObject *where_was G_GNUC_UNUSED)
+{
+    VirtViewerDisplay *display = VIRT_VIEWER_DISPLAY(data);
+
+    g_object_set_data(G_OBJECT(display), "virt-viewer-window", NULL);
+}
+
 static VirtViewerWindow *
 ensure_window_for_display(VirtViewerApp *self, VirtViewerDisplay *display)
 {
     gint nth = virt_viewer_display_get_nth(display);
     VirtViewerWindow *win = virt_viewer_app_get_nth_window(self, nth);
+
+    if (VIRT_VIEWER_IS_DISPLAY_VTE(display)) {
+        win = g_object_get_data(G_OBJECT(display), "virt-viewer-window");
+    }
+
     if (win == NULL) {
         GList *l = self->priv->windows;
 
@@ -912,12 +898,26 @@ ensure_window_for_display(VirtViewerApp *self, VirtViewerDisplay *display)
         }
 
         virt_viewer_window_set_display(win, display);
+        if (VIRT_VIEWER_IS_DISPLAY_VTE(display)) {
+            g_object_set_data(G_OBJECT(display), "virt-viewer-window", win);
+            g_object_weak_ref(G_OBJECT(win), window_weak_notify, display);
+        }
     }
     virt_viewer_app_set_window_subtitle(self, win, nth);
 
     return win;
 }
 
+static VirtViewerWindow *
+display_show_notebook_get_window(VirtViewerApp *self, VirtViewerDisplay *display)
+{
+    VirtViewerWindow *win = ensure_window_for_display(self, display);
+    VirtViewerNotebook *nb = virt_viewer_window_get_notebook(win);
+
+    virt_viewer_notebook_show_display(nb);
+    return win;
+}
+
 static void
 display_show_hint(VirtViewerDisplay *display,
                   GParamSpec *pspec G_GNUC_UNUSED,
@@ -945,9 +945,7 @@ display_show_hint(VirtViewerDisplay *display,
             virt_viewer_window_hide(win);
     } else {
         if (hint & VIRT_VIEWER_DISPLAY_SHOW_HINT_READY) {
-            win = ensure_window_for_display(self, display);
-            nb = virt_viewer_window_get_notebook(win);
-            virt_viewer_notebook_show_display(nb);
+            win = display_show_notebook_get_window(self, display);
             virt_viewer_window_show(win);
         } else {
             if (!self->priv->kiosk && win) {
@@ -967,8 +965,15 @@ virt_viewer_app_display_added(VirtViewerSession *session G_GNUC_UNUSED,
     gint nth;
 
     g_object_get(display, "nth-display", &nth, NULL);
-
     g_debug("Insert display %d %p", nth, display);
+
+    if (VIRT_VIEWER_IS_DISPLAY_VTE(display)) {
+        VirtViewerWindow *win = display_show_notebook_get_window(self, display);
+        virt_viewer_window_hide(win);
+        virt_viewer_app_update_menu_displays(self);
+        return;
+    }
+
     g_hash_table_insert(self->priv->displays, GINT_TO_POINTER(nth), g_object_ref(display));
 
     g_signal_connect(display, "notify::show-hint",
@@ -976,7 +981,6 @@ virt_viewer_app_display_added(VirtViewerSession *session G_GNUC_UNUSED,
     g_object_notify(G_OBJECT(display), "show-hint"); /* call display_show_hint */
 }
 
-
 static void virt_viewer_app_remove_nth_window(VirtViewerApp *self,
                                               gint nth)
 {
@@ -1564,10 +1568,6 @@ virt_viewer_app_get_property (GObject *object, guint property_id,
         g_value_set_boolean(value, virt_viewer_app_get_enable_accel(self));
         break;
 
-    case PROP_HAS_FOCUS:
-        g_value_set_boolean(value, priv->focused > 0);
-        break;
-
     case PROP_KIOSK:
         g_value_set_boolean(value, priv->kiosk);
         break;
@@ -1580,6 +1580,14 @@ virt_viewer_app_get_property (GObject *object, guint property_id,
         g_value_set_string(value, priv->uuid);
         break;
 
+    case PROP_VM_UI:
+        g_value_set_boolean(value, priv->vm_ui);
+        break;
+
+    case PROP_VM_RUNNING:
+        g_value_set_boolean(value, priv->vm_running);
+        break;
+
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -1634,6 +1642,14 @@ virt_viewer_app_set_property (GObject *object, guint property_id,
         virt_viewer_app_set_uuid_string(self, g_value_get_string(value));
         break;
 
+    case PROP_VM_UI:
+        priv->vm_ui = g_value_get_boolean(value);
+        break;
+
+    case PROP_VM_RUNNING:
+        priv->vm_running = g_value_get_boolean(value);
+        break;
+
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
     }
@@ -1731,7 +1747,7 @@ static void
 virt_viewer_app_init(VirtViewerApp *self)
 {
     GError *error = NULL;
-    self->priv = GET_PRIVATE(self);
+    self->priv = virt_viewer_app_get_instance_private(self);
 
     gtk_window_set_default_icon_name("virt-viewer");
 
@@ -1790,15 +1806,20 @@ virt_viewer_update_smartcard_accels(VirtViewerApp *self)
         sw_smartcard = FALSE;
     }
     if (sw_smartcard) {
+        gboolean r;
         g_debug("enabling smartcard shortcuts");
-        gtk_accel_map_change_entry("<virt-viewer>/file/smartcard-insert",
-                                   priv->insert_smartcard_accel_key,
-                                   priv->insert_smartcard_accel_mods,
-                                   TRUE);
-        gtk_accel_map_change_entry("<virt-viewer>/file/smartcard-remove",
-                                   priv->remove_smartcard_accel_key,
-                                   priv->remove_smartcard_accel_mods,
-                                   TRUE);
+        r = gtk_accel_map_change_entry("<virt-viewer>/file/smartcard-insert",
+                                       priv->insert_smartcard_accel_key,
+                                       priv->insert_smartcard_accel_mods,
+                                       TRUE);
+        if (!r)
+            g_warning("Unable to set hotkey for 'smartcard-insert' due to a conflict in GTK");
+        r = gtk_accel_map_change_entry("<virt-viewer>/file/smartcard-remove",
+                                       priv->remove_smartcard_accel_key,
+                                       priv->remove_smartcard_accel_mods,
+                                       TRUE);
+        if (!r)
+            g_warning("Unable to set hotkey for 'smartcard-remove' due to a conflict in GTK");
     } else {
         g_debug("disabling smartcard shortcuts");
         gtk_accel_map_change_entry("<virt-viewer>/file/smartcard-insert", 0, 0, TRUE);
@@ -1863,7 +1884,6 @@ virt_viewer_app_local_command_line (GApplication   *gapp,
 {
     VirtViewerApp *self = VIRT_VIEWER_APP(gapp);
     gboolean ret = FALSE;
-    gint argc = g_strv_length(*args);
     GError *error = NULL;
     GOptionContext *context = g_option_context_new(NULL);
     GOptionGroup *group = g_option_group_new("virt-viewer", NULL, NULL, gapp, NULL);
@@ -1882,7 +1902,7 @@ virt_viewer_app_local_command_line (GApplication   *gapp,
     g_option_context_add_group(context, spice_get_option_group());
 #endif
 
-    if (!g_option_context_parse(context, &argc, args, &error)) {
+    if (!g_option_context_parse_strv(context, args, &error)) {
         if (error != NULL) {
             g_printerr(_("%s\n"), error->message);
             g_error_free(error);
@@ -1913,8 +1933,6 @@ virt_viewer_app_class_init (VirtViewerAppClass *klass)
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
     GApplicationClass *g_app_class = G_APPLICATION_CLASS(klass);
 
-    g_type_class_add_private (klass, sizeof (VirtViewerAppPrivate));
-
     object_class->get_property = virt_viewer_app_get_property;
     object_class->set_property = virt_viewer_app_set_property;
     object_class->dispose = virt_viewer_app_dispose;
@@ -1997,15 +2015,6 @@ virt_viewer_app_class_init (VirtViewerAppClass *klass)
                                                          G_PARAM_READWRITE |
                                                          G_PARAM_STATIC_STRINGS));
 
-    g_object_class_install_property(object_class,
-                                    PROP_HAS_FOCUS,
-                                    g_param_spec_boolean("has-focus",
-                                                         "Has Focus",
-                                                         "Application has focus",
-                                                         FALSE,
-                                                         G_PARAM_READABLE |
-                                                         G_PARAM_STATIC_STRINGS));
-
     g_object_class_install_property(object_class,
                                     PROP_KIOSK,
                                     g_param_spec_boolean("kiosk",
@@ -2034,6 +2043,24 @@ virt_viewer_app_class_init (VirtViewerAppClass *klass)
                                                         G_PARAM_READABLE |
                                                         G_PARAM_WRITABLE |
                                                         G_PARAM_STATIC_STRINGS));
+
+    g_object_class_install_property(object_class,
+                                    PROP_VM_UI,
+                                    g_param_spec_boolean("vm-ui",
+                                                         "VM UI",
+                                                         "QEMU UI & behaviour",
+                                                         FALSE,
+                                                         G_PARAM_READWRITE |
+                                                         G_PARAM_STATIC_STRINGS));
+
+    g_object_class_install_property(object_class,
+                                    PROP_VM_RUNNING,
+                                    g_param_spec_boolean("vm-running",
+                                                         "VM running",
+                                                         "VM running",
+                                                         FALSE,
+                                                         G_PARAM_READWRITE |
+                                                         G_PARAM_STATIC_STRINGS));
 }
 
 void
@@ -2109,12 +2136,13 @@ virt_viewer_app_set_hotkeys(VirtViewerApp *self, const gchar *hotkeys_str)
             continue;
         }
 
+        gboolean status = TRUE;
         if (g_str_equal(*hotkey, "toggle-fullscreen")) {
-            gtk_accel_map_change_entry("<virt-viewer>/view/toggle-fullscreen", accel_key, accel_mods, TRUE);
+            status = gtk_accel_map_change_entry("<virt-viewer>/view/toggle-fullscreen", accel_key, accel_mods, TRUE);
         } else if (g_str_equal(*hotkey, "release-cursor")) {
-            gtk_accel_map_change_entry("<virt-viewer>/view/release-cursor", accel_key, accel_mods, TRUE);
+            status = gtk_accel_map_change_entry("<virt-viewer>/view/release-cursor", accel_key, accel_mods, TRUE);
         } else if (g_str_equal(*hotkey, "secure-attention")) {
-            gtk_accel_map_change_entry("<virt-viewer>/send/secure-attention", accel_key, accel_mods, TRUE);
+            status = gtk_accel_map_change_entry("<virt-viewer>/send/secure-attention", accel_key, accel_mods, TRUE);
         } else if (g_str_equal(*hotkey, "smartcard-insert")) {
             virt_viewer_set_insert_smartcard_accel(self, accel_key, accel_mods);
         } else if (g_str_equal(*hotkey, "smartcard-remove")) {
@@ -2122,6 +2150,8 @@ virt_viewer_app_set_hotkeys(VirtViewerApp *self, const gchar *hotkeys_str)
         } else {
             g_warning("Unknown hotkey command %s", *hotkey);
         }
+        if (!status)
+            g_warning("Unable to set hotkey for '%s' due to a conflict in GTK", *hotkey);
     }
     g_strfreev(hotkeys);
 
@@ -2232,19 +2262,21 @@ menu_display_visible_toggled_cb(GtkCheckMenuItem *checkmenuitem,
 {
     VirtViewerApp *self = virt_viewer_session_get_app(virt_viewer_display_get_session(display));
     gboolean visible = gtk_check_menu_item_get_active(checkmenuitem);
-    static gboolean reentering = FALSE;
     VirtViewerWindow *vwin;
 
-    if (reentering) /* do not reenter if I switch you back */
-        return;
-
-    reentering = TRUE;
-
     vwin = ensure_window_for_display(self, display);
-    visible = virt_viewer_app_window_set_visible(self, vwin, visible);
 
-    gtk_check_menu_item_set_active(checkmenuitem, /* will be toggled again */ !visible);
-    reentering = FALSE;
+    if (visible) {
+        virt_viewer_window_show(vwin);
+    } else {
+        if (virt_viewer_app_get_n_windows_visible(self) > 1) {
+            virt_viewer_window_hide(vwin);
+        } else {
+            virt_viewer_app_maybe_quit(self, vwin);
+            /* the last item remains active, doesn't matter if we quit */
+            gtk_check_menu_item_set_active(checkmenuitem, TRUE);
+        }
+    }
 
     virt_viewer_session_update_displays_geometry(virt_viewer_display_get_session(display));
 }
@@ -2343,6 +2375,26 @@ window_update_menu_displays_cb(gpointer value,
         tmp = tmp->next;
     }
 
+    for (tmp = self->priv->windows; tmp; tmp = tmp->next) {
+        VirtViewerWindow *win = VIRT_VIEWER_WINDOW(tmp->data);
+        VirtViewerDisplay *display = virt_viewer_window_get_display(win);
+
+        if (VIRT_VIEWER_IS_DISPLAY_VTE(display)) {
+            gchar *name = NULL;
+            GtkWidget *item;
+
+            g_object_get(display, "name", &name, NULL);
+            item = gtk_check_menu_item_new_with_label(name);
+            g_free(name);
+
+            virt_viewer_signal_connect_object(G_OBJECT(item), "toggled",
+                G_CALLBACK(menu_display_visible_toggled_cb), display, 0);
+            gtk_check_menu_item_set_active(GTK_CHECK_MENU_ITEM(item),
+                gtk_widget_get_visible(GTK_WIDGET(virt_viewer_window_get_window(win))));
+            gtk_menu_shell_append(submenu, item);
+        }
+    }
+
     gtk_widget_show_all(GTK_WIDGET(submenu));
     g_list_free(keys);
 }
