@@ -34,10 +34,12 @@
 #include "virt-viewer-util.h"
 #include "virt-viewer-session-spice.h"
 #include "virt-viewer-display-spice.h"
+#include "virt-viewer-display-vte.h"
 #include "virt-viewer-auth.h"
 
-G_DEFINE_TYPE (VirtViewerSessionSpice, virt_viewer_session_spice, VIRT_VIEWER_TYPE_SESSION)
-
+#if SPICE_GTK_CHECK_VERSION(0,36,0)
+#define WITH_QMP_PORT 1
+#endif
 
 struct _VirtViewerSessionSpicePrivate {
     GtkWindow *main_window;
@@ -51,10 +53,13 @@ struct _VirtViewerSessionSpicePrivate {
     guint pass_try;
     gboolean did_auto_conf;
     VirtViewerFileTransferDialog *file_transfer_dialog;
-
+    GError *disconnect_error;
+#ifdef WITH_QMP_PORT
+    SpiceQmpPort *qmp;
+#endif
 };
 
-#define VIRT_VIEWER_SESSION_SPICE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), VIRT_VIEWER_TYPE_SESSION_SPICE, VirtViewerSessionSpicePrivate))
+G_DEFINE_TYPE_WITH_PRIVATE (VirtViewerSessionSpice, virt_viewer_session_spice, VIRT_VIEWER_TYPE_SESSION)
 
 enum {
     PROP_0,
@@ -76,10 +81,13 @@ static void virt_viewer_session_spice_channel_new(SpiceSession *s,
 static void virt_viewer_session_spice_channel_destroy(SpiceSession *s,
                                                       SpiceChannel *channel,
                                                       VirtViewerSession *session);
+static void virt_viewer_session_spice_session_disconnected(SpiceSession *s,
+                                                           VirtViewerSessionSpice *session);
 static void virt_viewer_session_spice_smartcard_insert(VirtViewerSession *session);
 static void virt_viewer_session_spice_smartcard_remove(VirtViewerSession *session);
 static gboolean virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self);
 static void virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *self, GHashTable *monitors);
+static void virt_viewer_session_spice_vm_action(VirtViewerSession *self, gint action);
 
 static void virt_viewer_session_spice_clear_displays(VirtViewerSessionSpice *self)
 {
@@ -153,6 +161,7 @@ virt_viewer_session_spice_dispose(GObject *obj)
         gtk_widget_destroy(GTK_WIDGET(spice->priv->file_transfer_dialog));
         spice->priv->file_transfer_dialog = NULL;
     }
+    g_clear_error(&spice->priv->disconnect_error);
 
     G_OBJECT_CLASS(virt_viewer_session_spice_parent_class)->dispose(obj);
 }
@@ -259,8 +268,7 @@ virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
     dclass->apply_monitor_geometry = virt_viewer_session_spice_apply_monitor_geometry;
     dclass->can_share_folder = virt_viewer_session_spice_can_share_folder;
     dclass->can_retry_auth = virt_viewer_session_spice_can_retry_auth;
-
-    g_type_class_add_private(klass, sizeof(VirtViewerSessionSpicePrivate));
+    dclass->vm_action = virt_viewer_session_spice_vm_action;
 
     g_object_class_install_property(oclass,
                                     PROP_SPICE_SESSION,
@@ -286,7 +294,7 @@ virt_viewer_session_spice_class_init(VirtViewerSessionSpiceClass *klass)
 static void
 virt_viewer_session_spice_init(VirtViewerSessionSpice *self G_GNUC_UNUSED)
 {
-    self->priv = VIRT_VIEWER_SESSION_SPICE_GET_PRIVATE(self);
+    self->priv = virt_viewer_session_spice_get_instance_private(self);
 }
 
 static void
@@ -399,6 +407,8 @@ create_spice_session(VirtViewerSessionSpice *self)
                                       G_CALLBACK(virt_viewer_session_spice_channel_new), self, 0);
     virt_viewer_signal_connect_object(self->priv->session, "channel-destroy",
                                       G_CALLBACK(virt_viewer_session_spice_channel_destroy), self, 0);
+    virt_viewer_signal_connect_object(self->priv->session, "disconnected",
+                                      G_CALLBACK(virt_viewer_session_spice_session_disconnected), self, 0);
 
     usb_manager = spice_usb_device_manager_get(self->priv->session, NULL);
     if (usb_manager) {
@@ -456,6 +466,9 @@ virt_viewer_session_spice_close(VirtViewerSession *session)
 
     g_object_add_weak_pointer(G_OBJECT(self), (gpointer*)&self);
 
+#ifdef WITH_QMP_PORT
+    g_clear_object(&self->priv->qmp);
+#endif
     virt_viewer_session_spice_clear_displays(self);
 
     if (self->priv->session) {
@@ -501,21 +514,28 @@ fill_session(VirtViewerFile *file, SpiceSession *session)
     g_return_if_fail(VIRT_VIEWER_IS_FILE(file));
     g_return_if_fail(SPICE_IS_SESSION(session));
 
-    if (virt_viewer_file_is_set(file, "host")) {
-        gchar *val = virt_viewer_file_get_host(file);
-        g_object_set(G_OBJECT(session), "host", val, NULL);
+    if (virt_viewer_file_is_set(file, "unix-path")) {
+        gchar *val = virt_viewer_file_get_unix_path(file);
+        g_object_set(G_OBJECT(session), "unix-path", val, NULL);
         g_free(val);
-    }
+    } else {
+        if (virt_viewer_file_is_set(file, "host")) {
+            gchar *val = virt_viewer_file_get_host(file);
+            g_object_set(G_OBJECT(session), "host", val, NULL);
+            g_free(val);
+        }
 
-    if (virt_viewer_file_is_set(file, "port")) {
-        gchar *port = g_strdup_printf("%d", virt_viewer_file_get_port(file));
-        g_object_set(G_OBJECT(session), "port", port, NULL);
-        g_free(port);
-    }
-    if (virt_viewer_file_is_set(file, "tls-port")) {
-        gchar *tls_port = g_strdup_printf("%d", virt_viewer_file_get_tls_port(file));
-        g_object_set(G_OBJECT(session), "tls-port", tls_port, NULL);
-        g_free(tls_port);
+        if (virt_viewer_file_is_set(file, "port")) {
+            gchar *port = g_strdup_printf("%d", virt_viewer_file_get_port(file));
+            g_object_set(G_OBJECT(session), "port", port, NULL);
+            g_free(port);
+        }
+
+        if (virt_viewer_file_is_set(file, "tls-port")) {
+            gchar *tls_port = g_strdup_printf("%d", virt_viewer_file_get_tls_port(file));
+            g_object_set(G_OBJECT(session), "tls-port", tls_port, NULL);
+            g_free(tls_port);
+        }
     }
 
     if (virt_viewer_file_is_set(file, "username")) {
@@ -775,7 +795,7 @@ virt_viewer_session_spice_main_channel_event(SpiceChannel *channel,
         virt_viewer_session_spice_channel_destroy(NULL, channel, session);
         break;
     default:
-        g_warning("unhandled spice main channel event: %d", event);
+        g_warning("unhandled spice main channel event: %u", event);
         break;
     }
 
@@ -921,9 +941,9 @@ virt_viewer_session_spice_display_monitors(SpiceChannel *channel,
             !display_is_in_fullscreen_mode(self, VIRT_VIEWER_DISPLAY(display))) {
             g_debug("display %d should not be enabled, disabling",
                     virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)) + 1);
-            spice_main_set_display_enabled(virt_viewer_session_spice_get_main_channel(self),
-                                           virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)),
-                                           FALSE);
+            spice_main_channel_update_display_enabled(virt_viewer_session_spice_get_main_channel(self),
+                                                      virt_viewer_display_get_nth(VIRT_VIEWER_DISPLAY(display)),
+                                                      FALSE, TRUE);
             disabled = TRUE;
         }
 
@@ -951,20 +971,220 @@ on_new_file_transfer(SpiceMainChannel *channel G_GNUC_UNUSED,
                                               task);
 }
 
+static void
+spice_port_write_finished(GObject *source_object,
+                          GAsyncResult *res,
+                          gpointer dup)
+{
+    SpicePortChannel *port = SPICE_PORT_CHANNEL(source_object);
+    GError *err = NULL;
+
+    spice_port_channel_write_finish(port, res, &err);
+    if (err) {
+        g_warning("Spice port write failed: %s", err->message);
+        g_error_free(err);
+    }
+    g_free(dup);
+}
+
+static void
+spice_vte_commit(SpicePortChannel *port, const char *text,
+                 guint size, gpointer user_data G_GNUC_UNUSED)
+{
+    void *dup = g_memdup(text, size);
+
+    /* note: spice-gtk queues write */
+    spice_port_channel_write_async(port, dup, size,
+                                   NULL, spice_port_write_finished, dup);
+}
+
+static void
+spice_port_data(VirtViewerDisplayVte *vte, gpointer data, int size,
+                SpicePortChannel *port G_GNUC_UNUSED)
+{
+    virt_viewer_display_vte_feed(vte, data, size);
+}
+
+static const char *
+port_name_to_vte_name(const char *name)
+{
+    if (g_str_equal(name, "org.qemu.console.serial.0"))
+        return _("Serial");
+    else if (g_str_equal(name, "org.qemu.monitor.hmp.0"))
+        return _("QEMU human monitor");
+    else if (g_str_equal(name, "org.qemu.console.debug.0"))
+        return _("QEMU debug console");
+
+    return NULL;
+}
+
+static void
+virt_viewer_session_spice_vm_action(VirtViewerSession *sess G_GNUC_UNUSED,
+                                    gint action G_GNUC_UNUSED)
+{
+#ifdef WITH_QMP_PORT
+    VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(sess);
+
+    switch (action) {
+    case VIRT_VIEWER_SESSION_VM_ACTION_QUIT:
+        action = SPICE_QMP_PORT_VM_ACTION_QUIT;
+        break;
+    case VIRT_VIEWER_SESSION_VM_ACTION_RESET:
+        action = SPICE_QMP_PORT_VM_ACTION_RESET;
+        break;
+    case VIRT_VIEWER_SESSION_VM_ACTION_POWER_DOWN:
+        action = SPICE_QMP_PORT_VM_ACTION_POWER_DOWN;
+        break;
+    case VIRT_VIEWER_SESSION_VM_ACTION_PAUSE:
+        action = SPICE_QMP_PORT_VM_ACTION_PAUSE;
+        break;
+    case VIRT_VIEWER_SESSION_VM_ACTION_CONTINUE:
+        action = SPICE_QMP_PORT_VM_ACTION_CONTINUE;
+        break;
+    default:
+        g_return_if_reached();
+    }
+
+    spice_qmp_port_vm_action_async(self->priv->qmp, action, NULL, NULL, NULL);
+#endif
+}
+
+#ifdef WITH_QMP_PORT
+static void
+set_vm_running(VirtViewerSessionSpice *self, gboolean running)
+{
+    g_object_set(virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self)),
+                 "vm-running", running, NULL);
+}
+
+static void
+query_status_cb(GObject *source_object G_GNUC_UNUSED,
+                GAsyncResult *res, gpointer user_data)
+{
+    VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(user_data);
+    SpiceQmpStatus *status;
+    gboolean running = TRUE;
+    GError *error = NULL;
+
+    status = spice_qmp_port_query_status_finish(self->priv->qmp, res, &error);
+    if (!status) {
+        g_warning("failed to query VM status: %s", error->message);
+        g_error_free(error);
+        return;
+    }
+
+    if (g_str_equal(status->status, "paused")) {
+        running = FALSE;
+    }
+
+    set_vm_running(self, running);
+
+    spice_qmp_status_unref(status);
+}
+
+static void qmp_ready_cb(VirtViewerSessionSpice *self,
+                         GParamSpec *pspec G_GNUC_UNUSED,
+                         GObject *object G_GNUC_UNUSED)
+{
+    spice_qmp_port_query_status_async(self->priv->qmp, NULL, query_status_cb, self);
+}
+
+static void qmp_event_cb(VirtViewerSessionSpice *self, const gchar *event,
+                         void *data G_GNUC_UNUSED, GObject *object G_GNUC_UNUSED)
+{
+    g_debug("QMP event %s", event);
+
+    if (g_str_equal(event, "STOP")) {
+        set_vm_running(self, FALSE);
+    } else if (g_str_equal(event, "RESUME")) {
+        set_vm_running(self, TRUE);
+    }
+}
+#endif /* WITH_QMP_PORT */
+
+static void
+spice_port_opened(SpiceChannel *channel, GParamSpec *pspec G_GNUC_UNUSED,
+                  VirtViewerSessionSpice *self)
+{
+    SpicePortChannel *port = SPICE_PORT_CHANNEL(channel);
+    int id;
+    gchar *name = NULL;
+    gboolean opened = FALSE;
+    const char *vte_name;
+    GtkWidget *vte;
+
+    g_object_get(G_OBJECT(port),
+                 "channel-id", &id,
+                 "port-name", &name,
+                 "port-opened", &opened,
+                 NULL);
+
+    g_return_if_fail(name != NULL);
+    g_debug("port#%d %s: %s", id, name, opened ? "opened" : "closed");
+    vte_name = port_name_to_vte_name(name);
+
+#ifdef WITH_QMP_PORT
+    if (g_str_equal(name, "org.qemu.monitor.qmp.0")) {
+        if (opened) {
+            g_return_if_fail(!self->priv->qmp);
+
+            g_object_set(virt_viewer_session_get_app(VIRT_VIEWER_SESSION(self)),
+                         "vm-ui", TRUE, NULL);
+
+            self->priv->qmp = spice_qmp_port_get(port);
+            virt_viewer_signal_connect_object(self->priv->qmp, "notify::ready",
+                                              G_CALLBACK(qmp_ready_cb), self, G_CONNECT_SWAPPED);
+            virt_viewer_signal_connect_object(self->priv->qmp, "event",
+                                              G_CALLBACK(qmp_event_cb), self, G_CONNECT_SWAPPED);
+        } else {
+            g_clear_object(&self->priv->qmp);
+        }
+        g_free(name);
+        return;
+    }
+#endif
+
+    g_free(name);
+    vte = g_object_get_data(G_OBJECT(port), "virt-viewer-vte");
+    if (vte) {
+        if (opened)
+            return;
+
+        g_object_set_data(G_OBJECT(port), "virt-viewer-vte", NULL);
+        virt_viewer_session_remove_display(VIRT_VIEWER_SESSION(self), VIRT_VIEWER_DISPLAY(vte));
+        g_object_unref(vte);
+
+    } else if (opened) {
+        if (!vte_name)
+            return;
+
+        vte = virt_viewer_display_vte_new(VIRT_VIEWER_SESSION(self), vte_name);
+        g_object_set_data(G_OBJECT(port), "virt-viewer-vte", g_object_ref_sink(vte));
+        virt_viewer_session_add_display(VIRT_VIEWER_SESSION(self), VIRT_VIEWER_DISPLAY(vte));
+        virt_viewer_signal_connect_object(vte, "commit",
+                                          G_CALLBACK(spice_vte_commit), port, G_CONNECT_SWAPPED);
+        virt_viewer_signal_connect_object(port, "port-data",
+                                          G_CALLBACK(spice_port_data), vte, G_CONNECT_SWAPPED);
+    }
+}
+
 static void
 virt_viewer_session_spice_channel_new(SpiceSession *s,
                                       SpiceChannel *channel,
                                       VirtViewerSession *session)
 {
     VirtViewerSessionSpice *self = VIRT_VIEWER_SESSION_SPICE(session);
-    int id;
+    int id, type;
 
     g_return_if_fail(self != NULL);
 
     virt_viewer_signal_connect_object(channel, "open-fd",
                                       G_CALLBACK(virt_viewer_session_spice_channel_open_fd_request), self, 0);
 
-    g_object_get(channel, "channel-id", &id, NULL);
+    g_object_get(channel,
+                 "channel-id", &id,
+                 "channel-type", &type,
+                 NULL);
 
     g_debug("New spice channel %p %s %d", channel, g_type_name(G_OBJECT_TYPE(channel)), id);
 
@@ -1013,6 +1233,13 @@ virt_viewer_session_spice_channel_new(SpiceSession *s,
             virt_viewer_session_set_has_usbredir(session, TRUE);
     }
 
+    /* the port channel object is also sub-classed for webdav... */
+    if (SPICE_IS_PORT_CHANNEL(channel) && type == SPICE_CHANNEL_PORT) {
+        virt_viewer_signal_connect_object(channel, "notify::port-opened",
+                                          G_CALLBACK(spice_port_opened), self, 0);
+        spice_channel_connect(channel);
+    }
+
     self->priv->channel_count++;
 }
 
@@ -1055,7 +1282,7 @@ virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self)
         return FALSE;
     }
 
-    spice_main_set_display_enabled(cmain, -1, FALSE);
+    spice_main_channel_update_display_enabled(cmain, -1, FALSE, TRUE);
 
     initial_displays = virt_viewer_app_get_initial_displays(app);
     ndisplays = g_list_length(initial_displays);
@@ -1063,11 +1290,11 @@ virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self)
     displays = g_hash_table_new_full(g_direct_hash, g_direct_equal, NULL, g_free);
 
     for (l = initial_displays; l != NULL; l = l->next) {
-        GdkRectangle* rect = g_new0(GdkRectangle, 1);;
         gint j = virt_viewer_app_get_initial_monitor_for_display(app, GPOINTER_TO_INT(l->data));
         if (j == -1)
             continue;
 
+        GdkRectangle* rect = g_new0(GdkRectangle, 1);;
         gdk_screen_get_monitor_geometry(screen, j, rect);
         g_hash_table_insert(displays, l->data, rect);
     }
@@ -1079,19 +1306,27 @@ virt_viewer_session_spice_fullscreen_auto_conf(VirtViewerSessionSpice *self)
         GdkRectangle *rect = value;
         gint j = GPOINTER_TO_INT(key);
 
-        spice_main_set_display(cmain, j, rect->x, rect->y, rect->width, rect->height);
-        spice_main_set_display_enabled(cmain, j, TRUE);
+        spice_main_channel_update_display(cmain, j, rect->x, rect->y, rect->width, rect->height, TRUE);
+        spice_main_channel_update_display_enabled(cmain, j, TRUE, TRUE);
         g_debug("Set SPICE display %d to (%d,%d)-(%dx%d)",
                   j, rect->x, rect->y, rect->width, rect->height);
     }
     g_list_free(initial_displays);
     g_hash_table_unref(displays);
 
-    spice_main_send_monitor_config(cmain);
+    spice_main_channel_send_monitor_config(cmain);
     self->priv->did_auto_conf = TRUE;
     return TRUE;
 }
 
+static void
+virt_viewer_session_spice_session_disconnected(G_GNUC_UNUSED SpiceSession *s,
+                                               VirtViewerSessionSpice *self)
+{
+    GError *error = self->priv->disconnect_error;
+    g_signal_emit_by_name(self, "session-disconnected", error ? error->message : NULL);
+}
+
 static void
 virt_viewer_session_spice_channel_destroy(G_GNUC_UNUSED SpiceSession *s,
                                           SpiceChannel *channel,
@@ -1119,6 +1354,17 @@ virt_viewer_session_spice_channel_destroy(G_GNUC_UNUSED SpiceSession *s,
         g_object_set_data(G_OBJECT(channel), "virt-viewer-displays", NULL);
     }
 
+    if (SPICE_IS_PORT_CHANNEL(channel)) {
+        VirtViewerDisplayVte *vte = g_object_get_data(G_OBJECT(channel), "virt-viewer-vte");
+        g_debug("zap port channel (#%d)", id);
+        if (vte) {
+            g_object_set_data(G_OBJECT(channel), "virt-viewer-vte", NULL);
+            virt_viewer_session_remove_display(VIRT_VIEWER_SESSION(self), VIRT_VIEWER_DISPLAY(vte));
+            g_object_unref(vte);
+        }
+
+    }
+
     if (SPICE_IS_PLAYBACK_CHANNEL(channel) && self->priv->audio) {
         g_debug("zap audio channel");
         self->priv->audio = NULL;
@@ -1130,10 +1376,12 @@ virt_viewer_session_spice_channel_destroy(G_GNUC_UNUSED SpiceSession *s,
         if (self->priv->usbredir_channel_count == 0)
             virt_viewer_session_set_has_usbredir(session, FALSE);
     }
-
-    self->priv->channel_count--;
-    if (self->priv->channel_count == 0)
-        g_signal_emit_by_name(self, "session-disconnected", error ? error->message : NULL);
+    if (error) {
+        g_warning("Channel error: %s", error->message);
+        if (self->priv->disconnect_error == NULL) {
+            self->priv->disconnect_error = g_error_copy(error);
+        }
+    }
 }
 
 VirtViewerSession *
@@ -1180,8 +1428,8 @@ virt_viewer_session_spice_apply_monitor_geometry(VirtViewerSession *session, GHa
         gint i = GPOINTER_TO_INT(key);
         GdkRectangle* rect = value;
 
-        spice_main_set_display(self->priv->main_channel, i, rect->x,
-                               rect->y, rect->width, rect->height);
+        spice_main_channel_update_display(self->priv->main_channel, i, rect->x,
+                                          rect->y, rect->width, rect->height, TRUE);
     }
 }
 
