@@ -42,6 +42,9 @@
 #include "virt-viewer-app.h"
 #include "virt-viewer-util.h"
 #include "virt-viewer-timed-revealer.h"
+#include "virt-viewer-display-vte.h"
+
+#include "remote-viewer-iso-list-dialog.h"
 
 #define ZOOM_STEP 10
 
@@ -49,6 +52,9 @@
 void virt_viewer_window_menu_view_zoom_out(GtkWidget *menu, VirtViewerWindow *self);
 void virt_viewer_window_menu_view_zoom_in(GtkWidget *menu, VirtViewerWindow *self);
 void virt_viewer_window_menu_view_zoom_reset(GtkWidget *menu, VirtViewerWindow *self);
+void virt_viewer_window_menu_machine_reset(GtkWidget *menu, VirtViewerWindow *self);
+void virt_viewer_window_menu_machine_powerdown(GtkWidget *menu, VirtViewerWindow *self);
+void virt_viewer_window_menu_machine_pause(GtkWidget *menu, VirtViewerWindow *self);
 gboolean virt_viewer_window_delete(GtkWidget *src, void *dummy, VirtViewerWindow *self);
 void virt_viewer_window_menu_file_quit(GtkWidget *src, VirtViewerWindow *self);
 void virt_viewer_window_guest_details_response(GtkDialog *dialog, gint response_id, gpointer user_data);
@@ -62,6 +68,7 @@ void virt_viewer_window_menu_file_smartcard_insert(GtkWidget *menu, VirtViewerWi
 void virt_viewer_window_menu_file_smartcard_remove(GtkWidget *menu, VirtViewerWindow *self);
 void virt_viewer_window_menu_view_release_cursor(GtkWidget *menu, VirtViewerWindow *self);
 void virt_viewer_window_menu_preferences_cb(GtkWidget *menu, VirtViewerWindow *self);
+void virt_viewer_window_menu_change_cd_activate(GtkWidget *menu, VirtViewerWindow *self);
 
 
 /* Internal methods */
@@ -73,11 +80,6 @@ static GtkMenu* virt_viewer_window_get_keycombo_menu(VirtViewerWindow *self);
 static void virt_viewer_window_get_minimal_dimensions(VirtViewerWindow *self, guint *width, guint *height);
 static gint virt_viewer_window_get_minimal_zoom_level(VirtViewerWindow *self);
 
-G_DEFINE_TYPE (VirtViewerWindow, virt_viewer_window, G_TYPE_OBJECT)
-
-#define GET_PRIVATE(o)                                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE ((o), VIRT_VIEWER_TYPE_WINDOW, VirtViewerWindowPrivate))
-
 enum {
     PROP_0,
     PROP_WINDOW,
@@ -114,6 +116,8 @@ struct _VirtViewerWindowPrivate {
     gboolean initial_zoom_set;
 };
 
+G_DEFINE_TYPE_WITH_PRIVATE (VirtViewerWindow, virt_viewer_window, G_TYPE_OBJECT)
+
 static void
 virt_viewer_window_get_property (GObject *object, guint property_id,
                                  GValue *value, GParamSpec *pspec)
@@ -216,7 +220,32 @@ rebuild_combo_menu(GObject    *gobject G_GNUC_UNUSED,
     menu = GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-send"));
     gtk_menu_item_set_submenu(GTK_MENU_ITEM(menu),
                               GTK_WIDGET(virt_viewer_window_get_keycombo_menu(self)));
-    gtk_widget_set_sensitive(menu, (self->priv->display != NULL));
+}
+
+static void
+vm_ui_changed(GObject    *gobject G_GNUC_UNUSED,
+              GParamSpec *pspec G_GNUC_UNUSED,
+              gpointer    user_data)
+{
+    VirtViewerWindow *self = user_data;
+    gboolean vm_ui;
+
+    g_object_get(G_OBJECT(self->priv->app), "vm-ui", &vm_ui, NULL);
+    gtk_widget_set_visible(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-machine")), vm_ui);
+}
+
+static void
+vm_running_changed(GObject    *gobject G_GNUC_UNUSED,
+                   GParamSpec *pspec G_GNUC_UNUSED,
+                   gpointer    user_data)
+{
+    VirtViewerWindow *self = user_data;
+    GtkCheckMenuItem *check = GTK_CHECK_MENU_ITEM(gtk_builder_get_object(self->priv->builder, "menu-vm-pause"));
+    gboolean running;
+
+    g_object_get(G_OBJECT(self->priv->app), "vm-running", &running, NULL);
+
+    gtk_check_menu_item_set_active(check, !running);
 }
 
 static void
@@ -229,6 +258,10 @@ virt_viewer_window_constructed(GObject *object)
 
     g_signal_connect(priv->app, "notify::enable-accel",
                      G_CALLBACK(rebuild_combo_menu), object);
+    g_signal_connect(priv->app, "notify::vm-ui",
+                     G_CALLBACK(vm_ui_changed), object);
+    g_signal_connect(priv->app, "notify::vm-running",
+                     G_CALLBACK(vm_running_changed), object);
     rebuild_combo_menu(NULL, NULL, object);
 }
 
@@ -237,8 +270,6 @@ virt_viewer_window_class_init (VirtViewerWindowClass *klass)
 {
     GObjectClass *object_class = G_OBJECT_CLASS (klass);
 
-    g_type_class_add_private (klass, sizeof (VirtViewerWindowPrivate));
-
     object_class->get_property = virt_viewer_window_get_property;
     object_class->set_property = virt_viewer_window_set_property;
     object_class->dispose = virt_viewer_window_dispose;
@@ -299,7 +330,7 @@ virt_viewer_window_init (VirtViewerWindow *self)
     GtkWidget *vbox;
     GSList *accels;
 
-    self->priv = GET_PRIVATE(self);
+    self->priv = virt_viewer_window_get_instance_private(self);
     priv = self->priv;
 
     priv->fullscreen_monitor = -1;
@@ -310,11 +341,6 @@ virt_viewer_window_init (VirtViewerWindow *self)
 
     priv->builder = virt_viewer_util_load_ui("virt-viewer.ui");
 
-    gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-send")), FALSE);
-    gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-view-zoom")), FALSE);
-    gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-file-screenshot")), FALSE);
-    gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-preferences")), FALSE);
-
     gtk_builder_connect_signals(priv->builder, self);
 
     priv->accel_group = GTK_ACCEL_GROUP(gtk_builder_get_object(priv->builder, "accelgroup"));
@@ -382,27 +408,69 @@ virt_viewer_window_get_real_zoom_level(VirtViewerWindow *self)
     return round((double) NORMAL_ZOOM_LEVEL * allocation.width / width);
 }
 
+G_MODULE_EXPORT void
+virt_viewer_window_menu_machine_reset(GtkWidget *menu G_GNUC_UNUSED,
+                                      VirtViewerWindow *self)
+{
+    virt_viewer_session_vm_action(virt_viewer_app_get_session(self->priv->app),
+                                  VIRT_VIEWER_SESSION_VM_ACTION_RESET);
+}
+
+G_MODULE_EXPORT void
+virt_viewer_window_menu_machine_powerdown(GtkWidget *menu G_GNUC_UNUSED,
+                                      VirtViewerWindow *self)
+{
+    virt_viewer_session_vm_action(virt_viewer_app_get_session(self->priv->app),
+                                  VIRT_VIEWER_SESSION_VM_ACTION_POWER_DOWN);
+}
+
+G_MODULE_EXPORT void
+virt_viewer_window_menu_machine_pause(GtkWidget *menu G_GNUC_UNUSED,
+                                      VirtViewerWindow *self)
+{
+    gint action;
+
+    if (gtk_check_menu_item_get_active(GTK_CHECK_MENU_ITEM(menu)))
+        action = VIRT_VIEWER_SESSION_VM_ACTION_PAUSE;
+    else
+        action = VIRT_VIEWER_SESSION_VM_ACTION_CONTINUE;
+
+    virt_viewer_session_vm_action(virt_viewer_app_get_session(self->priv->app), action);
+}
+
 G_MODULE_EXPORT void
 virt_viewer_window_menu_view_zoom_out(GtkWidget *menu G_GNUC_UNUSED,
                                       VirtViewerWindow *self)
 {
-    virt_viewer_window_set_zoom_level(self,
-                                      virt_viewer_window_get_real_zoom_level(self) - ZOOM_STEP);
+    if (VIRT_VIEWER_IS_DISPLAY_VTE(self->priv->display)) {
+        virt_viewer_display_vte_zoom_out(VIRT_VIEWER_DISPLAY_VTE(self->priv->display));
+    } else {
+        virt_viewer_window_set_zoom_level(self,
+                                          virt_viewer_window_get_real_zoom_level(self) - ZOOM_STEP);
+    }
 }
 
 G_MODULE_EXPORT void
 virt_viewer_window_menu_view_zoom_in(GtkWidget *menu G_GNUC_UNUSED,
                                      VirtViewerWindow *self)
 {
-    virt_viewer_window_set_zoom_level(self,
-                                      virt_viewer_window_get_real_zoom_level(self) + ZOOM_STEP);
+    if (VIRT_VIEWER_IS_DISPLAY_VTE(self->priv->display)) {
+        virt_viewer_display_vte_zoom_in(VIRT_VIEWER_DISPLAY_VTE(self->priv->display));
+    } else {
+        virt_viewer_window_set_zoom_level(self,
+                                          virt_viewer_window_get_real_zoom_level(self) + ZOOM_STEP);
+    }
 }
 
 G_MODULE_EXPORT void
 virt_viewer_window_menu_view_zoom_reset(GtkWidget *menu G_GNUC_UNUSED,
                                         VirtViewerWindow *self)
 {
-    virt_viewer_window_set_zoom_level(self, NORMAL_ZOOM_LEVEL);
+    if (VIRT_VIEWER_IS_DISPLAY_VTE(self->priv->display)) {
+        virt_viewer_display_vte_zoom_reset(VIRT_VIEWER_DISPLAY_VTE(self->priv->display));
+    } else {
+        virt_viewer_window_set_zoom_level(self, NORMAL_ZOOM_LEVEL);
+    }
 }
 
 /* Kick GtkWindow to tell it to adjust to our new widget sizes */
@@ -411,10 +479,14 @@ virt_viewer_window_queue_resize(VirtViewerWindow *self)
 {
     VirtViewerWindowPrivate *priv = self->priv;
     GtkRequisition nat;
+    GtkWidget *child;
+    guint border;
 
+    border = gtk_container_get_border_width(GTK_CONTAINER(priv->window));
+    child = gtk_bin_get_child(GTK_BIN(priv->window));
     gtk_window_set_default_size(GTK_WINDOW(priv->window), -1, -1);
-    gtk_widget_get_preferred_size(priv->window, NULL, &nat);
-    gtk_window_resize(GTK_WINDOW(priv->window), nat.width, nat.height);
+    gtk_widget_get_preferred_size(child, NULL, &nat);
+    gtk_window_resize(GTK_WINDOW(priv->window), nat.width + border, nat.height + border);
 }
 
 static void
@@ -511,8 +583,11 @@ virt_viewer_window_enter_fullscreen(VirtViewerWindow *self, gint monitor)
 
     virt_viewer_window_menu_fullscreen_set_active(self, TRUE);
     gtk_widget_hide(menu);
-    gtk_widget_show(priv->toolbar);
-    virt_viewer_timed_revealer_force_reveal(priv->revealer, TRUE);
+
+    if (!priv->kiosk) {
+        gtk_widget_show(priv->toolbar);
+        virt_viewer_timed_revealer_force_reveal(priv->revealer, TRUE);
+    }
 
     if (priv->display) {
         virt_viewer_display_set_monitor(priv->display, monitor);
@@ -520,7 +595,19 @@ virt_viewer_window_enter_fullscreen(VirtViewerWindow *self, gint monitor)
     }
     virt_viewer_window_move_to_monitor(self);
 
-    gtk_window_fullscreen(GTK_WINDOW(priv->window));
+#if GTK_CHECK_VERSION(3, 18, 0)
+    if (monitor == -1) {
+        // just go fullscreen on the current monitor
+        gtk_window_fullscreen(GTK_WINDOW(priv->window));
+    } else {
+G_GNUC_BEGIN_IGNORE_DEPRECATIONS
+        gtk_window_fullscreen_on_monitor(GTK_WINDOW(priv->window),
+                                         gdk_screen_get_default(), monitor);
+G_GNUC_END_IGNORE_DEPRECATIONS
+    }
+#else
+        gtk_window_fullscreen(GTK_WINDOW(priv->window));
+#endif /* GTK_CHECK_VERSION */
 }
 
 #define MAX_KEY_COMBO 4
@@ -686,6 +773,45 @@ virt_viewer_window_get_keycombo_menu(VirtViewerWindow *self)
     return menu;
 }
 
+static void
+action_zoom_in(G_GNUC_UNUSED GSimpleAction *action,
+               G_GNUC_UNUSED GVariant *state,
+               gpointer user_data)
+{
+    virt_viewer_window_menu_view_zoom_in(NULL, VIRT_VIEWER_WINDOW(user_data));
+}
+
+static void
+action_zoom_out(G_GNUC_UNUSED GSimpleAction *action,
+                G_GNUC_UNUSED GVariant *state,
+                gpointer user_data)
+{
+    virt_viewer_window_menu_view_zoom_out(NULL, VIRT_VIEWER_WINDOW(user_data));
+}
+
+static void
+action_zoom_reset(G_GNUC_UNUSED GSimpleAction *action,
+                  G_GNUC_UNUSED GVariant *state,
+                  gpointer user_data)
+{
+    virt_viewer_window_menu_view_zoom_reset(NULL, VIRT_VIEWER_WINDOW(user_data));
+}
+
+/* Keep keypad_action_entries and keypad_action_accels in sync */
+static const GActionEntry keypad_action_entries[] = {
+    { .name = "zoom-in", .activate = action_zoom_in },
+    { .name = "zoom-out", .activate = action_zoom_out },
+    { .name = "zoom-reset", .activate = action_zoom_reset },
+};
+
+static const gchar *const keypad_action_accels[][2] = {
+    /* numpad keys are not handled automatically by gtk, see bgo#699823 */
+    {"<control>KP_Add", NULL},
+    {"<control>KP_Subtract", NULL},
+    {"<control>KP_0", NULL},
+};
+G_STATIC_ASSERT(G_N_ELEMENTS(keypad_action_entries) == G_N_ELEMENTS(keypad_action_accels));
+
 void
 virt_viewer_window_disable_modifiers(VirtViewerWindow *self)
 {
@@ -693,6 +819,7 @@ virt_viewer_window_disable_modifiers(VirtViewerWindow *self)
     VirtViewerWindowPrivate *priv = self->priv;
     GValue empty;
     GSList *accels;
+    guint i;
 
     if (!priv->accel_enabled)
         return;
@@ -719,6 +846,10 @@ virt_viewer_window_disable_modifiers(VirtViewerWindow *self)
                  "gtk-enable-mnemonics", FALSE,
                  NULL);
 
+    for (i = 0; i < G_N_ELEMENTS(keypad_action_entries); i++) {
+        g_action_map_remove_action(G_ACTION_MAP(priv->window), keypad_action_entries[i].name);
+    }
+
     priv->accel_enabled = FALSE;
 }
 
@@ -728,6 +859,7 @@ virt_viewer_window_enable_modifiers(VirtViewerWindow *self)
     GtkSettings *settings = gtk_settings_get_default();
     VirtViewerWindowPrivate *priv = self->priv;
     GSList *accels;
+    guint i;
 
     if (priv->accel_enabled)
         return;
@@ -748,6 +880,17 @@ virt_viewer_window_enable_modifiers(VirtViewerWindow *self)
                  "gtk-enable-mnemonics", priv->enable_mnemonics_save,
                  NULL);
 
+    g_action_map_add_action_entries(G_ACTION_MAP(priv->window),
+                                    keypad_action_entries, G_N_ELEMENTS(keypad_action_entries),
+                                    self);
+    for (i = 0; i < G_N_ELEMENTS(keypad_action_entries); i++) {
+        gchar *detailed_name = g_strdup_printf("win.%s", keypad_action_entries[i].name);
+        gtk_application_set_accels_for_action(GTK_APPLICATION(priv->app),
+                                              detailed_name,
+                                              keypad_action_accels[i]);
+        g_free(detailed_name);
+    }
+
     priv->accel_enabled = TRUE;
 }
 
@@ -844,7 +987,7 @@ static void add_if_writable (GdkPixbufFormat *data, GHashTable *formats)
     }
 }
 
-static GHashTable *init_image_formats(void)
+static GHashTable *init_image_formats(G_GNUC_UNUSED gpointer user_data)
 {
     GHashTable *format_map;
     GSList *formats = gdk_pixbuf_get_formats();
@@ -872,34 +1015,29 @@ static GdkPixbufFormat *get_image_format(const char *filename)
     return g_hash_table_lookup(image_formats_once.retval, ext);
 }
 
-static void
+static gboolean
 virt_viewer_window_save_screenshot(VirtViewerWindow *self,
-                                   const char *file)
+                                   const char *file,
+                                   GError **error)
 {
     VirtViewerWindowPrivate *priv = self->priv;
     GdkPixbuf *pix = virt_viewer_display_get_pixbuf(VIRT_VIEWER_DISPLAY(priv->display));
     GdkPixbufFormat *format = get_image_format(file);
+    gboolean result;
 
     if (format == NULL) {
-        g_debug("unknown file extension, falling back to png");
-        if (!g_str_has_suffix(file, ".png")) {
-            char *png_filename;
-            png_filename = g_strconcat(file, ".png", NULL);
-            gdk_pixbuf_save(pix, png_filename, "png", NULL,
-                            "tEXt::Generator App", PACKAGE, NULL);
-            g_free(png_filename);
-        } else {
-            gdk_pixbuf_save(pix, file, "png", NULL,
-                            "tEXt::Generator App", PACKAGE, NULL);
-        }
+        g_set_error(error, G_FILE_ERROR, G_FILE_ERROR_FAILED,
+                    _("Unable to determine image format for file '%s'"), file);
+        result = FALSE;
     } else {
         char *type = gdk_pixbuf_format_get_name(format);
         g_debug("saving to %s", type);
-        gdk_pixbuf_save(pix, file, type, NULL, NULL);
+        result = gdk_pixbuf_save(pix, file, type, error, NULL);
         g_free(type);
     }
 
     g_object_unref(pix);
+    return result;
 }
 
 G_MODULE_EXPORT void
@@ -924,13 +1062,18 @@ virt_viewer_window_menu_file_screenshot(GtkWidget *menu G_GNUC_UNUSED,
     image_dir = g_get_user_special_dir(G_USER_DIRECTORY_PICTURES);
     if (image_dir != NULL)
         gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER (dialog), image_dir);
-    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER (dialog), _("Screenshot"));
+    gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER (dialog), _("Screenshot.png"));
 
     if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT) {
         char *filename;
+        GError *error = NULL;
 
         filename = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER (dialog));
-        virt_viewer_window_save_screenshot(self, filename);
+        if (!virt_viewer_window_save_screenshot(self, filename, &error)) {
+            virt_viewer_app_simple_message_dialog(self->priv->app,
+                                                  error->message);
+            g_error_free(error);
+        }
         g_free(filename);
     }
 
@@ -1052,6 +1195,44 @@ virt_viewer_window_menu_help_about(GtkWidget *menu G_GNUC_UNUSED,
     g_object_unref(G_OBJECT(about));
 }
 
+#if HAVE_OVIRT
+static void
+iso_dialog_response(GtkDialog *dialog,
+                    gint response_id,
+                    gpointer user_data G_GNUC_UNUSED)
+{
+    if (response_id == GTK_RESPONSE_NONE)
+        return;
+
+    gtk_widget_destroy(GTK_WIDGET(dialog));
+}
+#endif
+
+G_MODULE_EXPORT void
+virt_viewer_window_menu_change_cd_activate(GtkWidget *menu G_GNUC_UNUSED,
+                                           VirtViewerWindow *self G_GNUC_UNUSED)
+{
+#if HAVE_OVIRT
+    VirtViewerWindowPrivate *priv = self->priv;
+    GtkWidget *dialog;
+    GObject *foreign_menu;
+
+    g_object_get(G_OBJECT(priv->app), "ovirt-foreign-menu", &foreign_menu, NULL);
+    dialog = remote_viewer_iso_list_dialog_new(GTK_WINDOW(priv->window), foreign_menu);
+    g_object_unref(foreign_menu);
+
+    if (!dialog)
+        dialog = gtk_message_dialog_new(GTK_WINDOW(priv->window),
+                                        GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT,
+                                        GTK_MESSAGE_ERROR,
+                                        GTK_BUTTONS_CLOSE,
+                                        _("Unable to connnect to oVirt"));
+
+    g_signal_connect(dialog, "response", G_CALLBACK(iso_dialog_response), NULL);
+    gtk_widget_show_all(dialog);
+    gtk_dialog_run(GTK_DIALOG(dialog));
+#endif
+}
 
 static void
 virt_viewer_window_toolbar_setup(VirtViewerWindow *self)
@@ -1241,13 +1422,24 @@ virt_viewer_window_set_menus_sensitive(VirtViewerWindow *self, gboolean sensitiv
     gtk_widget_set_sensitive(menu, sensitive);
 
     menu = GTK_WIDGET(gtk_builder_get_object(priv->builder, "menu-file-screenshot"));
-    gtk_widget_set_sensitive(menu, sensitive);
+    gtk_widget_set_sensitive(menu, sensitive &&
+                             VIRT_VIEWER_DISPLAY_CAN_SCREENSHOT(self->priv->display));
 
     menu = GTK_WIDGET(gtk_builder_get_object(priv->builder, "menu-view-zoom"));
     gtk_widget_set_sensitive(menu, sensitive);
 
-    menu = GTK_WIDGET(gtk_builder_get_object(priv->builder, "menu-send"));
+    menu = GTK_WIDGET(gtk_builder_get_object(priv->builder, "menu-machine"));
     gtk_widget_set_sensitive(menu, sensitive);
+
+    {
+        gboolean can_send = sensitive &&
+            VIRT_VIEWER_DISPLAY_CAN_SEND_KEYS(self->priv->display);
+
+        menu = GTK_WIDGET(gtk_builder_get_object(priv->builder, "menu-send"));
+        gtk_widget_set_sensitive(menu, can_send);
+
+        gtk_widget_set_sensitive(self->priv->toolbar_send_key, can_send);
+    }
 }
 
 static void
@@ -1328,10 +1520,7 @@ virt_viewer_window_set_display(VirtViewerWindow *self, VirtViewerDisplay *displa
         if (virt_viewer_display_get_enabled(display))
             virt_viewer_window_desktop_resize(display, self);
 
-        gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-preferences")), TRUE);
-        gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-view-zoom")), TRUE);
-        gtk_widget_set_sensitive(GTK_WIDGET(gtk_builder_get_object(self->priv->builder, "menu-send")), TRUE);
-        gtk_widget_set_sensitive(self->priv->toolbar_send_key, TRUE);
+        virt_viewer_window_set_menus_sensitive(self, TRUE);
     }
 }
 
@@ -1418,7 +1607,9 @@ virt_viewer_window_set_zoom_level(VirtViewerWindow *self, gint zoom_level)
 
     virt_viewer_display_set_zoom_level(VIRT_VIEWER_DISPLAY(priv->display), priv->zoomlevel);
 
-    virt_viewer_window_queue_resize(self);
+    if (!VIRT_VIEWER_IS_DISPLAY_VTE(priv->display)) {
+        virt_viewer_window_queue_resize(self);
+    }
 }
 
 gint virt_viewer_window_get_zoom_level(VirtViewerWindow *self)
